from datetime import timedelta

from pyasic_rs.asic_rs import Miner as _rs_Miner
from pyasic_rs.asic_rs import MinerModel as _rs_MinerModel
from pyasic_rs.asic_rs import HashAlgorithm as _rs_HashAlgorithm
from pyasic_rs.asic_rs import MinerFirmware as _rs_MinerFirmware
from pyasic_rs.asic_rs import MinerMake as _rs_MinerMake
from .data import MinerData, BoardData, HashRate, FanData, MinerMessage, PoolData


class Miner:
    def __init__(self, *, inner: _rs_Miner):
        self.__inner = inner

    def __repr__(self):
        return self.__inner.__repr__()

    @property
    def model(self) -> _rs_MinerModel:
        return self.__inner.model

    @property
    def make(self) -> _rs_MinerMake:
        return self.__inner.make

    @property
    def firmware(self) -> _rs_MinerMake:
        return self.__inner.firmware

    @property
    def algo(self) -> _rs_HashAlgorithm:
        return self.__inner.algo

    @property
    def expected_hashboards(self) -> int:
        return self.__inner.expected_hashboards

    @property
    def expected_chips(self) -> int:
        return self.__inner.expected_chips

    @property
    def expected_fans(self) -> int:
        return self.__inner.expected_fans

    async def get_data(self) -> MinerData:
        return MinerData.model_validate(await self.__inner.get_data())

    async def get_mac(self) -> str | None:
        return await self.__inner.get_mac()

    async def get_serial_number(self) -> str | None:
        return await self.__inner.get_serial_number()

    async def get_hostname(self) -> str | None:
        return await self.__inner.get_hostname()

    async def get_api_version(self) -> str | None:
        return await self.__inner.get_api_version()

    async def get_firmware_version(self) -> str | None:
        return await self.__inner.get_firmware_version()

    async def get_control_board_version(self) -> str | None:
        return await self.__inner.get_control_board_version()

    async def get_hashboards(self) -> list[BoardData]:
        return [BoardData.model_validate(b) for b in await self.__inner.get_hashboards()]

    async def get_hashrate(self) -> HashRate | None:
        inner = await self.__inner.get_hashrate()
        if inner is not None:
            return HashRate.model_validate(inner)
        return None

    async def get_expected_hashrate(self) -> HashRate | None:
        inner = await self.__inner.get_expected_hashrate()
        if inner is not None:
            return HashRate.model_validate(inner)
        return None

    async def get_fans(self) -> list[FanData]:
        return [FanData.model_validate(f) for f in await self.__inner.get_fans()]

    async def get_psu_fans(self) -> list[FanData]:
        return [FanData.model_validate(f) for f in await self.__inner.get_psu_fans()]

    async def get_fluid_temperature(self) -> float | None:
        return await self.__inner.get_fluid_temperature()

    async def get_wattage(self) -> float | None:
        return await self.__inner.get_wattage()

    async def get_wattage_limit(self) -> float | None:
        return await self.__inner.get_wattage_limit()

    async def get_light_flashing(self) -> bool | None:
        return await self.__inner.get_light_flashing()

    async def get_messages(self) -> list[MinerMessage]:
        return [MinerMessage.model_validate(m) for m in await self.__inner.get_messages()]

    async def get_uptime(self) -> timedelta | None:
        return await self.__inner.get_uptime()

    async def get_is_mining(self) -> bool | None:
        return await self.__inner.get_is_mining()

    async def get_pools(self) -> list[PoolData]:
        return [PoolData.model_validate(b) for b in await self.__inner.get_pools()]

    async def set_fault_light(self, fault: bool) -> bool | None:
        return await self.__inner.set_fault_light(fault)

    async def restart(self) -> bool | None:
        return await self.__inner.restart()

    async def pause(self, at_time: timedelta | int) -> bool | None:
        if isinstance(at_time, int):
            at_time = timedelta(seconds=at_time)
        return await self.__inner.pause(at_time)

    async def resume(self, at_time: timedelta | int) -> bool | None:
        if isinstance(at_time, int):
            at_time = timedelta(seconds=at_time)
        return await self.__inner.restart(at_time)



from datetime import timedelta
from enum import IntEnum
from ipaddress import IPv4Address
from typing import Annotated, Self

from pydantic import BaseModel, ConfigDict, BeforeValidator, field_serializer, model_serializer


class MinerHardware(BaseModel):
    model_config = ConfigDict(from_attributes=True)

    chips: int | None
    fans: int | None
    boards: int | None


class DeviceInfo(BaseModel):
    model_config = ConfigDict(from_attributes=True)

    make: Annotated[str, BeforeValidator(str)]
    model: Annotated[str, BeforeValidator(str)]
    hardware: MinerHardware
    firmware: Annotated[str, BeforeValidator(str)]
    algo: Annotated[str, BeforeValidator(str)]


class HashRateUnit(IntEnum):
    H = 1
    KH = H * 1000
    MH = KH * 1000
    GH = MH * 1000
    TH = GH * 1000
    PH = TH * 1000
    EH = PH * 1000
    ZH = EH * 1000

    default = TH

    @classmethod
    def from_asic_rs(cls, val):
        val = int(val)
        if val == 0:
            return cls.H
        if val == 1:
            return cls.KH
        if val == 2:
            return cls.MH
        if val == 3:
            return cls.GH
        if val == 4:
            return cls.TH
        if val == 5:
            return cls.PH
        if val == 6:
            return cls.EH
        if val == 7:
            return cls.ZH
        return cls.default

    def __str__(self):
        if self.value == self.H:
            return "H/s"
        if self.value == self.KH:
            return "KH/s"
        if self.value == self.MH:
            return "MH/s"
        if self.value == self.GH:
            return "GH/s"
        if self.value == self.TH:
            return "TH/s"
        if self.value == self.PH:
            return "PH/s"
        if self.value == self.EH:
            return "EH/s"
        if self.value == self.ZH:
            return "ZH/s"

    @classmethod
    def from_str(cls, value: str):
        if value == "H":
            return cls.H
        elif value == "KH":
            return cls.KH
        elif value == "MH":
            return cls.MH
        elif value == "GH":
            return cls.GH
        elif value == "TH":
            return cls.TH
        elif value == "PH":
            return cls.PH
        elif value == "EH":
            return cls.EH
        elif value == "ZH":
            return cls.ZH
        return cls.default

    def __repr__(self):
        return str(self)


class HashRate(BaseModel):
    model_config = ConfigDict(from_attributes=True)

    value: float
    unit: Annotated[HashRateUnit, BeforeValidator(HashRateUnit.from_asic_rs)]
    algo: str

    def __float__(self):
        return self.value

    def __format__(self, format_spec: str):
        formatted_value = format(self.value, format_spec)
        return f"{formatted_value} {self.unit}"

    @model_serializer
    def serialize_hashrate(self):
        return self.into_unit(unit=HashRateUnit.default).value

    def into_unit(self, unit: HashRateUnit) -> Self:
        return HashRate(
            value=(self.value / int(self.unit)) * int(unit),
            unit=unit,
            algo=self.algo
        )


class ChipData(BaseModel):
    model_config = ConfigDict(from_attributes=True)

    position: int
    hashrate: HashRate | None
    temperature: float | None
    voltage: float | None
    frequency: float | None
    tuned: bool | None
    working: bool | None


class BoardData(BaseModel):
    model_config = ConfigDict(from_attributes=True)

    position: int
    hashrate: HashRate | None
    expected_hashrate: HashRate | None
    board_temperature: float | None
    intake_temperature: float | None
    outlet_temperature: float | None
    expected_chips: int | None
    working_chips: int | None
    serial_number: str | None
    chips: list[ChipData]
    voltage: float | None
    frequency: float | None
    tuned: bool | None
    active: bool | None


class FanData(BaseModel):
    model_config = ConfigDict(from_attributes=True)

    position: int
    rpm: float | None


class PoolData(BaseModel):
    model_config = ConfigDict(from_attributes=True)

    position: int | None
    url: Annotated[str, BeforeValidator(str)] | None
    accepted_shares: int | None
    rejected_shares: int | None
    active: bool | None
    alive: bool | None
    user: str | None


class MinerMessage(BaseModel):
    model_config = ConfigDict(from_attributes=True)

    timestamp: int
    code: int
    message: str
    severity: Annotated[str, BeforeValidator(str)]


class MinerData(BaseModel):
    model_config = ConfigDict(from_attributes=True)

    schema_version: str
    timestamp: int
    ip: IPv4Address
    mac: str
    device_info: DeviceInfo
    serial_number: str | None
    hostname: str | None
    api_version: str | None
    firmware_version: str | None
    expected_hashboards: int | None
    hashboards: list[BoardData]
    hashrate: HashRate | None
    expected_hashrate: HashRate | None
    expected_chips: int | None
    total_chips: int | None
    expected_fans: int | None
    fans: list[FanData]
    psu_fans: list[FanData]
    average_temperature: float | None
    fluid_temperature: float | None
    wattage: float | None
    wattage_limit: float | None
    efficiency: float | None
    light_flashing: bool | None
    messages: list[MinerMessage]
    uptime: timedelta | None
    is_mining: bool
    pools: list[PoolData]

    @field_serializer("uptime")
    def serialize_uptime(self, uptime: timedelta, _info) -> float:
        return uptime.total_seconds()



from typing import Self, AsyncIterable

from pyasic_rs.asic_rs import MinerFactory as _rs_MinerFactory
from .miner import Miner


class MinerFactory:
    def __init__(self, *, inner: _rs_MinerFactory = _rs_MinerFactory()):
        self.__inner = inner

    def with_subnet(self, subnet: str) -> Self:
        self.__inner.with_subnet(subnet)
        return self

    @classmethod
    def from_subnet(cls, subnet: str) -> Self:
        return cls(inner=_rs_MinerFactory.from_subnet(subnet))

    def with_octets(self, octet_1: int | str, octet_2: int | str, octet_3: int | str, octet_4: int | str) -> Self:
        self.__inner.with_octets(octet_1, octet_2, octet_3, octet_4)
        return self

    @classmethod
    def from_octets(cls, octet_1: int | str, octet_2: int | str, octet_3: int | str, octet_4: int | str) -> Self:
        return cls(inner=_rs_MinerFactory.from_octets(str(octet_1), str(octet_2), str(octet_3), str(octet_4)))

    def with_range(self, ip_range: str) -> Self:
        self.__inner.with_range(ip_range)
        return self

    @classmethod
    def from_range(cls, ip_range: str) -> Self:
        return cls(inner=_rs_MinerFactory.from_range(ip_range))

    async def get_miner(self, ip: str) -> Miner | None:
        base = await self.__inner.get_miner(ip)
        if base is not None:
            return Miner(inner=base)
        return None

    async def scan(self) -> list[Miner]:
        bases = await self.__inner.scan()
        return [Miner(inner=m) for m in filter(lambda x: x is not None, bases)]

    def scan_stream(self) -> AsyncIterable[Miner]:
        return self.__inner.scan_stream()

    def scan_stream_with_ip(self) -> AsyncIterable[Miner]:
        return self.__inner.scan_stream_with_ip()
